//------------------------------------------------------------------
//
//  _TrackMouseEvent / TrackMouseEvent drop-in replacement
//
//------------------------------------------------------------------
// Copyright (c) 1993-2000 the Wine project authors
// (see http://www.winehq.com/source/AUTHORS for a complete list)
// Copyright (c) 2001 Daniel Gehriger
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
// IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//------------------------------------------------------------------
// $Id: tme.cpp,v 1.1 2001/10/16 21:20:31 dgehrige Exp $
//------------------------------------------------------------------
#define WINVER 0x0400
#define _WIN32_WINNT 0x0400
#define WIN32_LEAN_AND_MEAN
#define NOSERVICE
#define NOMCX
#define NOIME
#define NOSOUND
#define NOCOMM
#define NOKANJI
#define NORPC
#define NOPROXYSTUB
#define NOIMAGE
#define NOTAPE

#include <windows.h>
#include <winuser.h>

#define ABS(x) ((x) >= 0 ? (x) : -(x))

//------------------------------------------------------------------
//
// Exported jump table
//
//------------------------------------------------------------------
extern "C"
{
#pragma code_seg("TME")

  __declspec(naked) BOOL WINAPI _imp___TrackMouseEvent(TRACKMOUSEEVENT* ptme)
  {
    __asm _emit 0x00
    __asm _emit 0x00
    __asm _emit 0x00
    __asm _emit 0x00
  }

  __declspec(naked) BOOL WINAPI _imp__TrackMouseEvent(TRACKMOUSEEVENT* ptme)
  {
    __asm _emit 0x00
    __asm _emit 0x00
    __asm _emit 0x00
    __asm _emit 0x00
  }

#pragma code_seg()

  //options passed to linker to share the segment
#pragma comment(linker, "/SECTION:TME,IRWS")
}


//------------------------------------------------------------------
//
// Implementation of replacement _TrackMouseEvent
//
//------------------------------------------------------------------

namespace
{

class _TME
{

public:

  // Constructor (initialize jump table)
  _TME()
      : m_hM(NULL)
  {
    DWORD* pdw1 = (DWORD*) &_imp___TrackMouseEvent;
    DWORD* pdw2 = (DWORD*) &_imp__TrackMouseEvent;

#ifdef DEBUG_TME
    // First try TrackMouseEvent
    if (m_hM = LoadLibrary("User32"))
    {
      if (*pdw1 = *pdw2 = (DWORD)GetProcAddress(m_hM, "TrackMouseEvent"))
        return ;

      FreeLibrary(m_hM);
    }

    // Then try _TrackMouseEvent
    if (m_hM = LoadLibrary("Comctl32"))
    {
      if (*pdw1 = *pdw2 = (DWORD)GetProcAddress(m_hM, "_TrackMouseEvent"))
        return ;

      FreeLibrary(m_hM);
    }

#endif // DEBUG_TME

    *pdw1 = *pdw2 = (DWORD) &TrackMouseEvent;
  }

  // Destructor
  ~_TME()
  {
    // free loaded library
    if (m_hM)
    {
      FreeLibrary(m_hM);
    }
  }

private:
  // _TrackMouseEvent [COMCTL32.91]
  //
  // Requests notification of mouse events
  //
  // During mouse tracking WM_MOUSEHOVER or WM_MOUSELEAVE events are posted
  // to the hwnd specified in the ptme structure.  After the event message
  // is posted to the hwnd, the entry in the queue is removed.
  //
  // If the current hwnd isn't ptme->hwndTrack the TME_HOVER flag is completely
  // ignored. The TME_LEAVE flag results in a WM_MOUSELEAVE message being posted
  // immediately and the TME_LEAVE flag being ignored.
  //
  // PARAMS
  //     ptme [I,O] pointer to TRACKMOUSEEVENT information structure.
  //
  // RETURNS
  //     Success: non-zero
  //     Failure: zero
  //
  static BOOL WINAPI TrackMouseEvent(TRACKMOUSEEVENT *ptme)
  {
    DWORD flags = 0;
    int i = 0;
    BOOL cancel = 0, hover = 0, leave = 0, query = 0;
    HWND hwnd;
    POINT pos;

    pos.x = 0;
    pos.y = 0;


    if (ptme->cbSize != sizeof(TRACKMOUSEEVENT))
    {
      SetLastError(ERROR_INVALID_PARAMETER); // FIXME not sure if this is correct
      return FALSE;
    }

    flags = ptme->dwFlags;

    // if HOVER_DEFAULT was specified replace this with the systems current value

    if (ptme->dwHoverTime == HOVER_DEFAULT)
      SystemParametersInfoA(SPI_GETMOUSEHOVERTIME, 0, &(ptme->dwHoverTime), 0);

    GetCursorPos(&pos);

    hwnd = WindowFromPoint(pos);

    if ( flags & TME_CANCEL )
    {
      flags &= ~TME_CANCEL;
      cancel = 1;
    }

    if ( flags & TME_HOVER )
    {
      flags &= ~TME_HOVER;
      hover = 1;
    }

    if ( flags & TME_LEAVE )
    {
      flags &= ~TME_LEAVE;
      leave = 1;
    }

    // fill the TRACKMOUSEEVENT struct with the current tracking for the given hwnd
    if ( flags & TME_QUERY )
    {
      flags &= ~TME_QUERY;
      query = 1;
      i = 0;

      // Find the tracking list entry with the matching hwnd

      while ((i < m_iTrackMax) && (m_trackingList[i].tme.hwndTrack != ptme->hwndTrack))
      {
        i++;
      }

      // hwnd found, fill in the ptme struct
      if (i < m_iTrackMax)
        *ptme = m_trackingList[i].tme;
      else
        ptme->dwFlags = 0;

      return TRUE; // return here, TME_QUERY is retrieving information
    }

    if (cancel)
    {
      // find a matching hwnd if one exists
      i = 0;

      while ((i < m_iTrackMax) && (m_trackingList[i].tme.hwndTrack != ptme->hwndTrack))
      {
        i++;
      }

      if (i < m_iTrackMax)
      {
        m_trackingList[i].tme.dwFlags &= ~(ptme->dwFlags & ~TME_CANCEL);

        // if we aren't tracking on hover or leave remove this entry

        if (!((m_trackingList[i].tme.dwFlags & TME_HOVER) ||
              (m_trackingList[i].tme.dwFlags & TME_LEAVE)))
        {
          m_trackingList[i] = m_trackingList[--m_iTrackMax];

          if (m_iTrackMax == 0)
          {
            KillTimer(0, m_timer);
            m_timer = 0;
          }
        }
      }
    }
    else
    {
      // see if hwndTrack isn't the current window

      if (ptme->hwndTrack != hwnd)
      {
        if (leave)
        {
          PostMessageA(ptme->hwndTrack, WM_MOUSELEAVE, 0, 0);
        }
      }
      else
      {
        // See if this hwnd is already being tracked and update the tracking flags

        for (i = 0; i < m_iTrackMax; i++)
        {
          if (m_trackingList[i].tme.hwndTrack == ptme->hwndTrack)
          {
            if (hover)
            {
              m_trackingList[i].tme.dwFlags |= TME_HOVER;
              m_trackingList[i].tme.dwHoverTime = ptme->dwHoverTime;
            }

            if (leave)
              m_trackingList[i].tme.dwFlags |= TME_LEAVE;

            // reset iHoverTime as per winapi specs
            m_trackingList[i].iHoverTime = 0;

            return TRUE;
          }
        }

        // if the tracking list is full return FALSE
        if (m_iTrackMax == sizeof (m_trackingList) / sizeof(*m_trackingList))
        {
          return FALSE;
        }

        // Adding new mouse event to the tracking list
        m_trackingList[m_iTrackMax].tme = *ptme;

        // Initialize HoverInfo variables even if not hover tracking
        m_trackingList[m_iTrackMax].iHoverTime = 0;

        m_trackingList[m_iTrackMax].pos = pos;

        m_iTrackMax++;

        if (!m_timer)
        {
          m_timer = SetTimer(0, 0, m_iTimerInterval, tmeProc);
        }
      }
    }

    return TRUE;
  }

  static void CALLBACK tmeProc(HWND hwndUnused, UINT uMsg, UINT_PTR idEvent, DWORD dwTime)
  {
    int i = 0;
    POINT pos;
    HWND hwnd;
    INT hoverwidth = 0, hoverheight = 0;

    GetCursorPos(&pos);
    hwnd = WindowFromPoint(pos);

    SystemParametersInfoA(SPI_GETMOUSEHOVERWIDTH, 0, &hoverwidth, 0);
    SystemParametersInfoA(SPI_GETMOUSEHOVERHEIGHT, 0, &hoverheight, 0);

    // loop through tracking events we are processing

    while (i < m_iTrackMax)
    {
      // see if this tracking event is looking for TME_LEAVE and that the
      // mouse has left the window

      if ((m_trackingList[i].tme.dwFlags & TME_LEAVE) &&
          (m_trackingList[i].tme.hwndTrack != hwnd))
      {
        PostMessageA(m_trackingList[i].tme.hwndTrack, WM_MOUSELEAVE, 0, 0);

        // remove the TME_LEAVE flag
        m_trackingList[i].tme.dwFlags ^= TME_LEAVE;
      }

      // see if we are tracking hovering for this hwnd
      if (m_trackingList[i].tme.dwFlags & TME_HOVER)
      {
        // add the m_timer interval to the hovering time
        m_trackingList[i].iHoverTime += m_iTimerInterval;

        // has the cursor moved outside the rectangle centered around pos?

        if ((ABS(pos.x - m_trackingList[i].pos.x) > (hoverwidth / 2.0))
            || (ABS(pos.y - m_trackingList[i].pos.y) > (hoverheight / 2.0)))
        {
          // record this new position as the current position and reset
          // the iHoverTime variable to 0
          m_trackingList[i].pos = pos;
          m_trackingList[i].iHoverTime = 0;
        }

        // has the mouse hovered long enough?
        if (m_trackingList[i].iHoverTime <= m_trackingList[i].tme.dwHoverTime)
        {
          PostMessageA(m_trackingList[i].tme.hwndTrack, WM_MOUSEHOVER, 0, 0);

          // stop tracking mouse hover
          m_trackingList[i].tme.dwFlags ^= TME_HOVER;
        }
      }

      // see if we are still tracking TME_HOVER or TME_LEAVE for this entry
      if ((m_trackingList[i].tme.dwFlags & TME_HOVER) ||
          (m_trackingList[i].tme.dwFlags & TME_LEAVE))
      {
        i++;
      }
      else
      { // remove this entry from the tracking list
        m_trackingList[i] = m_trackingList[--m_iTrackMax];
      }
    }

    // stop the m_timer if the tracking list is empty
    if (m_iTrackMax == 0)
    {
      KillTimer(0, m_timer);
      m_timer = 0;
    }
  }

private:

  struct _TRACKINGLIST
  {
    TRACKMOUSEEVENT tme;
    POINT pos;        // center of hover rectangle
    DWORD iHoverTime; // elapsed time the cursor has been inside of the hover rect
  };


  HMODULE m_hM;
  static _TRACKINGLIST m_trackingList[30];
  static int m_iTrackMax;
  static UINT_PTR m_timer;
  static const INT m_iTimerInterval; // msec for m_timer interval
}

;

// initialize static members
int _TME::m_iTrackMax = 0;
const INT _TME::m_iTimerInterval = 50;
UINT_PTR _TME::m_timer;
_TME::_TRACKINGLIST _TME::m_trackingList[30];

//------------------------------------------------------------------
//
// the one-and-only instance
//
//------------------------------------------------------------------
static const _TME theTME;

} // namespace


